/***
 * The grid template is the representation of the solution space
 * for the OPTIMIST. A grid can be n-dimensional, where
 * coordinates are specified via an n-array, or vector. Each
 * entry (referred by the cooridnates) contains a collection (set)
 * of grid elements. An element of a grid is a subclass of the
 * 
 * A grid has the following methods:
 * - insert(_T, unsigned int coord[n])
 *     insert an element at coord
 * - bool exist(GridElement *e, unsigned int coord[n])
 *     true if there is an element e' at coord such that 
 *     e == e', with defined equality relation
 * - pair<iterator begin, iterator end> elements(coord)
 *     return iterators  for the set of elements of the grid
 *     positioned at coord.
 */


#ifndef __GRID_HH__
#define __GRID_HH__

#include <vector>
#include <map>
#include <ext/hash_set>
#include <assert.h>

template< class _T, unsigned int N >
class grid {

public:
#ifdef __GNUC__
  typedef typename __gnu_cxx::hash_set< _T , __gnu_cxx::hash< _T >, 
	  std::equal_to< _T > >::iterator element_iterator;
#else
  typedef typename std::hash_set< _T , std::hash< _T >, 
	  std::equal_to< _T > >::iterator element_iterator;
#endif //__GNUC_
  typedef grid< _T, N - 1 > sub_grid_t;
  typedef typename std::map< unsigned int, sub_grid_t, 
			     std::less< unsigned int > > sub_grid_map_t;
  sub_grid_map_t sub_grid_map;

  //exist
  std::pair<element_iterator, bool>
  exist(_T _ge, std::vector< unsigned int > _coord) {
    unsigned int _x = _coord.back();
    typename sub_grid_map_t::iterator sg_iter = sub_grid_map.find(_x);
    if(sg_iter == sub_grid_map.end()) {
#ifdef  __GNUC__
      __gnu_cxx::hash_set< _T , __gnu_cxx::hash< _T >, std::equal_to< _T > > dummy;
#else
      std::hash_set< _T , std::hash< _T >, std::equal_to< _T > > dummy;
#endif //__GNUC__
      element_iterator ei_end = dummy.end();
      return std::make_pair(ei_end, false);
    }
    _coord.pop_back();
    sub_grid_t & tmp_sub_grid = (*sg_iter).second;
    return tmp_sub_grid.exist(_ge, _coord);
  }

  //insert
  std::pair<element_iterator, bool>
  insert(_T _ge, std::vector< unsigned int > _coord) {
    unsigned int _x = _coord.back();
    _coord.pop_back();
    sub_grid_t & tmp_sub_grid = sub_grid_map[_x];
    return tmp_sub_grid.insert(_ge, _coord);
  }

  //erase
  void erase(_T _ge, std::vector< unsigned int > _coord) {
    unsigned int _x = _coord.back();
    _coord.pop_back();
    sub_grid_t & tmp_sub_grid = sub_grid_map[_x];
    tmp_sub_grid.erase(_ge, _coord);
  }

  //elements
  std::pair< element_iterator, element_iterator >
  elements(std::vector< unsigned int > _coord) {
    unsigned int _x = _coord.back();
    typename sub_grid_map_t::iterator sg_iter = sub_grid_map.find(_x);
    if(sg_iter == sub_grid_map.end()) {
#ifdef __GNUC__
      __gnu_cxx::hash_set< _T , __gnu_cxx::hash< _T >, std::equal_to< _T > > dummy;
#else
      std::hash_set< _T , std::hash< _T >, std::equal_to< _T > > dummy;
#endif //__GNUC__
      element_iterator ei = dummy.begin(), ei_end = dummy.end();
      return std::make_pair(ei, ei_end);
    }
    _coord.pop_back();
    sub_grid_t & tmp_sub_grid = (*sg_iter).second;
    return tmp_sub_grid.elements(_coord);
  }

  //empty
  bool empty(std::vector< unsigned int > _coord) {
    unsigned int _x = _coord.back();
    typename sub_grid_map_t::iterator sg_iter = sub_grid_map.find(_x);
    if(sg_iter == sub_grid_map.end()) {
      return true;
    }
    _coord.pop_back();
    sub_grid_t & tmp_sub_grid = (*sg_iter).second;
    return tmp_sub_grid.empty(_coord);
  }

  //clear
  void clear(std::vector< unsigned int > _coord) {
    unsigned int _x = _coord.back();
    _coord.pop_back();
    sub_grid_t & tmp_sub_grid = sub_grid_map[_x]; //FIXME
    tmp_sub_grid.clear(_coord);
  }
};

template< class _T >
class grid< _T, 0 > {

public:
#ifdef __GNUC__
  typedef typename __gnu_cxx::hash_set< _T, __gnu_cxx::hash< _T >, 
	  std::equal_to< _T > > element_set_t;
#else
  typedef typename std::hash_set< _T, std::hash< _T >, 
	  std::equal_to< _T > > element_set_t;
#endif //__GNUC__
  typedef typename element_set_t::iterator element_iterator;
  element_set_t elems;

  //exist
  std::pair<element_iterator, bool>
  exist(_T _ge, std::vector< unsigned int > _coord) {
    assert(_coord.size() == 0);
    element_iterator ei; //, ei_end = elems.end();
    if((ei = elems.find(_ge)) != elems.end()) {
      return std::make_pair(ei, true);
    }
    //for(ei = elems.begin(); ei != ei_end; ++ei) {
    //This relays that the grid contains pointers (solution?)
    //  if(*_ge == *(*ei)) {
    //	return std::make_pair(ei, true);
    //  }
    //}
    return std::make_pair(elems.end(), false);
  }

  //insert
  std::pair<element_iterator, bool>
  insert(_T _ge, std::vector< unsigned int > _coord) {
    assert(_coord.size() == 0);
    return elems.insert(_ge);
  }

  //erase
  void erase(_T _ge, std::vector< unsigned int > _coord) {
    assert(_coord.size() == 0);
    element_iterator ei = elems.find(_ge); //, ei_end;
    if(ei != elems.end()) {
      elems.erase(ei);
    }
    //for(ei = elems.begin(); ei != elems.end(); ++ei) {
      //This relays that the grid contains pointers (solution?)
      //it should be elems.erase(_ge) - where _ge is the key
      //if(*_ge == *(*ei)) {
      //elems.erase(ei);
      //break;
      //}
    //}
  }

  //elements
  std::pair< element_iterator, element_iterator >
  elements(std::vector< unsigned int > _coord) {
    assert(_coord.size() == 0);
    element_iterator ei = elems.begin(), ei_end = elems.end();
    return std::make_pair(ei, ei_end);
  }

  //empty
  bool empty(std::vector< unsigned int > _coord) {
    assert(_coord.size() == 0);
    return elems.empty();
  }

  //clear elements
  void clear(std::vector< unsigned int > _coord) {
    assert(_coord.size() == 0);
    elems.clear();
  }

};

#endif //__GRID_HH__
