#include "ls.hh"
#include "tp.hh"
#include "grid.hh"

ListSchedule::ListSchedule(Architecture *_architecture)
  : architecture(_architecture)
{
}

ListSchedule::~ListSchedule()
{
}

Schedule *ListSchedule::schedule(DAG & _dag, Flags & _flags)
{
  ZSet zset;
  Matcher matcher(architecture);
  NodeAttr *node_attr = get_init_node_attr(zset, _dag);
  TimeProfile *init_time_profile = new TimeProfile(architecture->get_reserv_tab_size());
  SNode *current_snode = 
    new SNode(node_attr, init_time_profile, architecture);

  vertex_set_t::iterator zi, zi_end;
  NodeAttr *current_node_attr = current_snode->get_node_attr();
  zi = current_node_attr->z.begin();
  zi_end = current_node_attr->z.end();
  SNode *partial_snode = NULL;

  while(current_node_attr->z.size() != 0) {

    partial_snode = find_ls_match(current_snode, matcher, zset, _dag);

    if(partial_snode == NULL) {
      std::cout << *current_snode << std::endl;
      bool found = false;
      //No match occured, try transfers.
      //transfer space
      grid< SNode *, 1 > transfer_space;
      std::vector< unsigned int > coord(1);
      coord[0] = 0;
      bool transfered_occured = false;

      transfer_space.insert(current_snode, coord);
      std::set< SNode * >::iterator gi, gi_end;
      do {
	for(tie(gi, gi_end) = transfer_space.elements(coord); gi != gi_end; ++gi) {
	  do {
	    SpaceProfile::iterator spi = current_snode->get_space_profile()->begin(),
	      spi_end = current_snode->get_space_profile()->end();
	    transfered_occured = false;

	    for( ; spi != spi_end; ++spi) {
	      std::set< int > & rcs = spi->second;
	      vertex_id u = spi->first;
	      for(std::set< int >::iterator src_iter = rcs.begin(); src_iter != rcs.end();
		  ++src_iter) {
		std::pair< bool, std::vector< std::pair< int, TargetInstruction * > > > dests =
		  architecture->get_transfer_destinations(*src_iter);
		if(!dests.first) {
		  continue;
		}
		std::vector< std::pair< int, TargetInstruction * > >::iterator dti_iter =
		  dests.second.begin(), dti_iter_end = dests.second.end();
		for( ; dti_iter != dti_iter_end; ++dti_iter) {
		  if(rcs.find(dti_iter->first) == rcs.end()) {
		    SNode *transfer_snode = new SNode(*current_snode);

		    //node u is not in residence class of the destination of the transfer instruction
		    //add transfer instruction
		    node_map_t dummy_node_map;
		    pattern_t dummy_pattern;

		    transfer_snode->insert(vertex_set_t(), u, _dag, dummy_node_map, dummy_pattern,
					   dti_iter->second, architecture);
		    std::pair< std::set< SNode *>::iterator, bool > ep;
		    ep = transfer_space.exist(transfer_snode, coord);
		    if(ep.second) {
		      //Already in the transfer space
		      delete transfer_snode;
		    } else {
		      //If a match, we stop!
		      if((partial_snode = find_ls_match(transfer_snode, matcher, zset, _dag)) != NULL) {
			//FIXME (free space from the transfer space) - delete on all nodes
			found = true;
		      }
		      transfered_occured = true;
		      transfer_space.insert(transfer_snode, coord);
		    }
		  }
		}
	      }
	    }
	  } while(transfered_occured && !found);
	}
	coord[0]++;
      } while(!transfer_space.empty(coord) && !found);
    }

    if(partial_snode == NULL) {
      std::cerr << "\nFailed to find a schedule with enhanced list scheduling!\n\n";
      assert(false);
    }
    delete current_snode;
    current_snode = partial_snode;
    current_node_attr = current_snode->get_node_attr();
  }

  Schedule *s = partial_snode->get_schedule();
  return s;
}


NodeAttr *ListSchedule::get_init_node_attr(ZSet & _zset, DAG & _dag)
{
  graph_traits< DAG >::edge_iterator ei, ei_end;
  graph_traits< DAG >::vertex_iterator vi, vi_end;
  indeg_t indeg;
  vertex_set_t z;
	
  //All vertices start with indegree 0
  for(tie(vi, vi_end) = vertices(_dag); vi != vi_end; ++vi) {
    indeg[*vi] = 0;
  }
  //Initial indegrees
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id v_id = target(*ei, _dag);
    indeg[v_id]++;
  }	
  //Initial z
  indeg_t::iterator ii;
  for(ii = indeg.begin(); ii != indeg.end(); ++ii) {
    if(indeg[ii->first] == 0) {
      z.insert(ii->first);
    }
  }
	
  vertex_set_t alive;
	
  //Initial set of alive is empty if we only consider basic block level
  alive.clear();
  NodeAttr *node_attr = _zset.insert(z, indeg, alive);
  return node_attr;
}

NodeAttr *ListSchedule::select(const vertex_set_t & _chi, 
			       const NodeAttr & _node_attr,
			       ZSet & _zset, DAG & _dag)
{
  vertex_set_t z(_node_attr.z);
  indeg_t indeg(_node_attr.indeg);
  vertex_set_t alive;

  vertex_set_t::const_iterator chi_iter = _chi.begin();
  for(; chi_iter != _chi.end(); ++chi_iter) {
    vertex_id u_id = *chi_iter;
    //successors of v_id are decremented
    successor_iterator pi, pi_end;
    for(tie(pi, pi_end) = successors(u_id, _dag);
	pi != pi_end; ++pi) {
      vertex_id v_id = *pi;
      indeg[v_id]--;
      if(indeg[v_id] == 0) {
	z.insert(v_id);
      }
    }
  }
  //Remove chi nodes from z
  for(chi_iter = _chi.begin(); chi_iter != _chi.end(); ++chi_iter) {
    assert(indeg[*chi_iter] == 0);
    z.erase(*chi_iter);
  }
  //Alive set
  graph_traits< DAG >::edge_iterator ei, ei_end;
  for(tie(ei, ei_end) = edges(_dag); ei != ei_end; ++ei) {
    vertex_id u_id = source(*ei, _dag),
      v_id = target(*ei, _dag);
    if((indeg[u_id] == 0) && (z.find(u_id) == z.end())) {
      if((indeg[v_id] != 0) || (z.find(v_id) != z.end())) {
	//Target in zero-indegree set
	alive.insert(u_id);
      }
    }			
  }
	
  NodeAttr *node_attr = _zset.insert(z, indeg, alive);
  return node_attr;
}

SNode *ListSchedule::find_ls_match(SNode *_current_snode, Matcher & _matcher, 
				   ZSet & _zset, DAG & _dag)
{
  NodeAttr *current_node_attr = _current_snode->get_node_attr();
  vertex_set_t::iterator zi, zi_end;
  zi = current_node_attr->z.begin();
  zi_end = current_node_attr->z.end();
  SNode *partial_snode = NULL;

  for( ; zi != zi_end; ++zi) {
    vertex_id u_id = *zi;
    Matcher::match_map_iter mi, mi_end;
    //For all match y covering (u_id, current_enode, & chi, & instr)
    for(tie(mi, mi_end) = _matcher.matches(u_id, *_current_snode, _dag);
	mi != mi_end; ++mi) {

      std::pair< pattern_t *, std::vector< iso_map_s > > & p = mi->second;
      
      for(std::vector< iso_map_s >::iterator iso_iter = p.second.begin();
	  iso_iter != p.second.end(); ++iso_iter) {

	vertex_set_t chi;
	chi.clear();
	for(node_map_t::iterator vi = (iso_iter->iso_map).begin(); 
	    vi != (iso_iter->iso_map).end(); ++vi) {
	  chi.insert(vi->second);
	}

	NodeAttr *succ_node_attr = select(chi, *current_node_attr, _zset, _dag);
	SNode *succ_snode = new SNode(*_current_snode);
	succ_snode->set_node_attr(succ_node_attr);
	std::pair< bool, int > ip = 
	  succ_snode->insert(chi, u_id, _dag, iso_iter->arg_map, *(p.first), 
			     mi->first, architecture);
	if(!ip.first) {
	  delete succ_snode;
	  continue;
	}

	if(partial_snode) {
	  if(partial_snode->get_tau() > succ_snode->get_tau()) {
	    //We check only partial time optimal
	    delete partial_snode;
	    partial_snode = succ_snode;
	  }
	} else { 
	  partial_snode = succ_snode;
	}
      }
    }
  }
  return partial_snode;
}
